////////////////////////////////////////////////////////////
//
// Filename:	tsbn.cc
// Description:	Implementation file for the TSBN class.
//              
// Copyright (c) 1998 Xiaojuan FENG
//
////////////////////////////////////////////////////////////

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>

#include "tsbn.hh"
#include "util.hh"
#include "node.hh"
#include "const.hh"

////////////////////////////////////////////////////////////
//
// TSBN::TSBN(s_int	nl,    
//            s_int*	nrows, 
//            s_int*	ncols, 
//            s_int*	np,     
//            s_int*	nc, 
//            s_int	nv,
//            float**	ve, 
//            float*	prob)
// 
// Description: The constructor of class: TSBN for virtual 
// evidences. The parameters are:
// nl	-- Num. of levels in the TSBN.
// nrows-- Num. of rows for each level.
// ncols-- Num. of cols for each level.
// np   -- Num. of parents.
// nc   -- Num. of children.
// nv   -- Num. of classes.
// ve   -- Pointer to virtual evidence matrix.
// prob -- Pointer to the prior probability array.
// 
////////////////////////////////////////////////////////////
/*
TSBN::TSBN(s_int	nl, 
           s_int*	nrows, 
           s_int*	ncols,
           s_int*	np,     
           s_int*	nc, 
           s_int	nv,
           float**	ve, 
           float*	prob)
{
 
 // initial the numbers
 num_levels = nl;
 num_rows = nrows;
 num_cols = ncols;
 num_parents = np;
 num_children = nc;
 num_classes = nv;
 nitemlambda = 0;
 nitempi = 0;
 virtual_evidence = ve;
 prior_prob = prob;
 evidence_type = VIRTUAL;
 scale_d = 0;
 // calculate the number of nodes
 num_nodes = GetNumNode();
 // calcaulate the numbers of nodes in each level
 GetNodeLevel();
 // allocation the nodes 
 tree = new NodeTrain2 [num_nodes];
 
 EstablishRelation(); 
 // allocation for stacks
 nitemstack = nodes_levels[num_levels-1]; 
               
 // Think more carefully !!!
 rlambda = new int [nitemstack];
 rpi = new int [nitemstack];

 
} // end of  TSBN::TSBN


*/

////////////////////////////////////////////////////////////
// TSBN::TSBN(s_int	nl, 
//            s_int*	nrows, 
//            s_int*	ncols,
//            s_int*	np, 
//            s_int*	nc, 
//            s_int	nv,
//            s_int*	labels, 
//            float*	priors,
//            float***	probs)
// 
// Description: The constructor of class: TSBN for training 
// network from label image (hard-evidences).
// The parameters are:
// nl	 -- Num. of levels in the TSBN.
// nrows -- Num. of rows for each level.
// ncols -- Num. of cols for each level.
// np    -- Num. of parents.
// nc    -- Num. of children.
// nv    -- Num. of classes.
// labes -- Pointer to hard evidence array.
// priors-- Pointer to the prior probability array.
// probs -- Pointer to the CPTs matrix (3D)
//      
////////////////////////////////////////////////////////////
/*
TSBN::TSBN(s_int	nl, 
           s_int*	nrows, 
           s_int*	ncols,
           s_int*	np, 
           s_int*	nc, 
           s_int	nv,
           s_int*	labels, 
           float*	priors,
           float***	probs)
{
            
 // initial the numbers
 num_levels = nl;
 num_rows = nrows;
 num_cols = ncols;
 num_parents = np;
 num_children = nc;
 num_classes = nv;
 nitemlambda = 0;
 nitempi = 0;
 hard_evidence = labels;
 prior_prob = priors;
 condp = probs;
 evidence_type = HARD;
 scale_d = 0;
 // calculate the number of nodes
 num_nodes = GetNumNode();
 // calcaulate the numbers of nodes in each level
 GetNodeLevel();
 // allocation the nodes 
 tree = new NodeTrain2 [num_nodes];
 
 EstablishRelation(); 
 // allocation for stacks
 nitemstack = nodes_levels[num_levels-1]; 
               
 // Think more carefully !!!
 rlambda = new int [nitemstack];
 rpi = new int [nitemstack];
 
} // end of  TSBN::TSBN


*/
////////////////////////////////////////////////////////////
// TSBN::TSBN
// 
// The constructor of class: TSBN 
// for training network from label-image (hard-evidence)
// 
// Modifications:
// 	08/01/98
//      
////////////////////////////////////////////////////////////
TSBN::TSBN(s_int  nl, 
           s_int* nrows, 
           s_int* ncols,
           s_int  ncpt,
           s_int* np, 
           s_int* nc, 
           s_int  nv)
{
            
 // initial the numbers
 num_levels = nl;
 num_rows = nrows;
 num_cols = ncols;
 num_cpts = ncpt;
 num_parents = np;
 num_children = nc;
 num_classes = nv;
 nitemlambda = 0;
 nitempi = 0;

 evidence_type = HARD;
 scale_d = 0;
 // calculate the number of nodes
 num_nodes = GetNumNode();
 // calcaulate the numbers of nodes in each level
 GetNodeLevel();
 // allocation the nodes 
 //tree1 = new NodeTrain2 [num_nodes];  // Temp:  13/01/99 Think deeply!!
 tree = new NodeTrain3 [num_nodes];  // Temp:  13/01/99 Think deeply!!
 
 EstablishRelation(); 
 // allocation for stacks
 nitemstack = nodes_levels[num_levels-1]; 
               
 // Think more carefully !!!
 rlambda = new int [nitemstack];
 rpi = new int [nitemstack];
 
} // end of  TSBN::TSBN


////////////////////////////////////////////////////////////
// TSBN::TSBN
// 
// The constructor of class: TSBN 
// for calculating the coding cost
// 
// Modifications:
// 	12/01/99
//      
////////////////////////////////////////////////////////////
TSBN::TSBN(s_int  nl, 
           s_int* nrows, 
           s_int* ncols,
           s_int  ncpt,
           s_int* np, 
           s_int* nc, 
           s_int  nv,
           char*  type )
{
            
 // initial the numbers
 num_levels = nl;
 num_rows = nrows;
 num_cols = ncols;
 num_cpts = ncpt;
 num_parents = np;
 num_children = nc;
 num_classes = nv;
 nitemlambda = 0;
 nitempi = 0;

 evidence_type = HARD;
 scale_d = 0;
 // calculate the number of nodes
 num_nodes = GetNumNode();
 // calcaulate the numbers of nodes in each level
 GetNodeLevel();
 // allocation the nodes 
 if (strcmp(type,"code") == 0)  
    tree = new NodeTrain3 [num_nodes];
 
 EstablishRelation(); 
 // allocation for stacks
 nitemstack = nodes_levels[num_levels-1]; 
               
 // Think more carefully !!!
 rlambda = new int [nitemstack];
 rpi = new int [nitemstack];
 
} // end of  TSBN::TSBN




////////////////////////////////////////////////////////////
//
// TSBN::~TSBN()
// Description: The destructor of class: TSBN
//
////////////////////////////////////////////////////////////
TSBN::~TSBN()
{
  //cout << "Destructor of class: TSBN.\n";
}


////////////////////////////////////////////////////////////
//
// TSBN::GetNumNode()
// Description: To calculate the number of nodes in TSBN
//
////////////////////////////////////////////////////////////
int TSBN::GetNumNode()
{
 int	i, n, nnodes;
 
 n = 1;
 nnodes = 1;
 for (i=0; i<num_levels; ++i) {
     n *= num_children[i];
     nnodes += n;
 }
 return nnodes;
} // end of TSBN::GetNumNode()


////////////////////////////////////////////////////////////
//
// TSBN::GetNodeLevel()
// Description: To calculate the number of nodes for each 
// level and the number of nodes above each level.
// The following two array variables are assigned values. 
// nodes_levels --
// sum_nodes_levels --
// 
////////////////////////////////////////////////////////////
void TSBN::GetNodeLevel()
{
 int	i, im1;

 nodes_levels = new int [num_levels];
 sum_nodes_levels = new int [num_levels];

 nodes_levels[0] = 1; 
 sum_nodes_levels[0] = 0;
 im1 = 0;
 for (i=1; i<num_levels; ++i) {
     sum_nodes_levels[i] = sum_nodes_levels[im1] + 
                           nodes_levels[im1];
     nodes_levels[i] = nodes_levels[im1]*num_children[im1];
     im1 = i;
 }
} // end of  TSBN::GetNodeLevel()


////////////////////////////////////////////////////////////
//
// TSBN::EstablishRelation()
// Description: To build up a (non)standard  
// quadtree-structured Belief network. 
//
////////////////////////////////////////////////////////////
void  TSBN::EstablishRelation()
{
 s_int      i, i2, j, j2, ki, kj;
 s_int 	    level1, leveln, lm1, la1;
 s_int      hrow, row, hcol, col;
 int        n, parent;

 // obtain the increasing times along rows and cols
 s_int* inrow = new s_int [num_levels];
 s_int* incol = new s_int [num_levels];
 for (i=1; i<num_levels; ++i) {
     inrow[i] = num_rows[i]/num_rows[i-1];
     incol[i] = num_cols[i]/num_cols[i-1];
 }
 // node in level 0
 n = 0;
 tree[n].parent = -1;
 tree[n].child1 = 1;
 ++n;
 // nodes in levels from 1 to (num_levels-1)
 level1 = num_levels-1;
 hrow = 1;
 for (leveln=1; leveln<num_levels; ++leveln) { 
     lm1 = leveln -1;
     la1 = leveln +1;
     row = num_rows[leveln];
     col = num_cols[leveln];
     hrow = num_rows[lm1];
     hcol = num_cols[lm1];
     for (i=0; i<hrow; ++i) {
         i2 = i*inrow[leveln];
         for (j=0; j<hcol; ++j) {
             j2 = j*incol[leveln];
             parent = sum_nodes_levels[lm1]+ i*hcol +j;
	     for (ki=0; ki<inrow[leveln]; ++ki) {
	         for (kj=0; kj<incol[leveln]; ++kj) {
                      tree[n].parent = parent;
                      if ( leveln < level1 )
                      tree[n].child1 = 
                      sum_nodes_levels[la1]+
                      (i2+ki)*col*num_children[leveln] + 
                      (j2+kj)*num_children[leveln];
                      else tree[n].child1 = -1;           
                      ++n;
                 }
            }
         }
     }
 }
 delete inrow;
 delete incol;

} // end of TSBN::EstablishRelation()


 
///////////////////////////////////////////////////////////
//
// void TSBN::SetCTP()
// Description: 
// Node: the indexes to condp is changed on 10/01/98
//        This function was written for the old 
//        structure of Node
//
//////////////////////////////////////////////////////////
/*
void TSBN::SetCTP(float** thetas)
{
 int	  i, j, k;
 //float    ***fmatrix(int, int, int);

 // allocation f a 3D matrix for cond_prob
 condp = fmatrix(num_classes,  num_classes, num_levels);
 // set the probs
 for (k=0; k<num_levels; ++k) {
     for (i=0; i<num_classes; ++i)
     for (j=0; j<num_classes; ++j)
         if ( i==j ) condp[i][j][k] = thetas[k][0];
         else condp[i][j][k] = thetas[k][1];
 }

} //end of TSBN::SetCTP() 
*/


///////////////////////////////////////////////////////////
//
// void TSBN::SetCPT1()
// Description: 
// For the case that there is only on CPT
// between any two adjacent levels, set the pointer of
// CTP index (condp_index) in each node to the 
// corresponding address in CPTs matrix (condp).
//
// The following variables are assigned values:
// condp_index in each node.
// 
///////////////////////////////////////////////////////////
void TSBN::SetCPT1(float*** probs)
{
 s_int 	i;
 s_int      n, le;

 prior_prob = new float [num_classes];

 for (i=0; i<num_classes; ++i)
     prior_prob[i] = probs[0][0][i];
 condp = probs;

 // nodes at level 1 to (num_levels-1)
 n = 1;
 for (le=1; le<num_levels; ++le) {
     for (i=0; i<nodes_levels[le]; ++i) 
         tree[sum_nodes_levels[le]+i].condp_index = n;
     ++n;
 }

} //end of TSBN::SetCPT1() 



//////////////////////////////////////////////////////////
//
// void SetCPT6()
// Description:  
// For the case that there are six or four 
// CPTs on the links from  root node to nodes on level 1,
// and only one CPT on the links between other levels, set 
// the pointer of  CTP index (condp_index) in each node to 
// the corresponding address in CPTs matrix (condp).
//
// The following variables are assigned values:
// condp_index in each node.
//
//////////////////////////////////////////////////////////
void TSBN::SetCPT6(float*** probs)
{
 s_int      i, le,qrow, qcol;
 int        n;

 prior_prob = new float [num_classes];

 for (i=0; i<num_classes; ++i)
     prior_prob[i] = probs[0][0][i];
 condp = probs;
 // nodes at level 1
 n = 1; 
 if (num_rows[num_levels-1] == num_cols[num_levels-1]) {
    qrow = 2;
    qcol = 2;
 }
 else {
    qrow = 2;
    qcol = 3;
 }
 for (i=1; i<=(qrow*qcol); ++i) 
     tree[i].condp_index = n++;
 // nodes at lower levels
 for (le=2; le<num_levels; ++le) {
     for (i=0; i<nodes_levels[le]; ++i) 
         tree[sum_nodes_levels[le]+i].condp_index = n;
     ++n;
 }
} //end of TSBN::SetCPT6() 



///////////////////////////////////////////////////////////
//
// void TSBN::GetCP()
// Description:
// 
// Node: the indexes to condp is changed on 10/01/98
//  ??
//////////////////////////////////////////////////////////
float TSBN::GetCP(int   child_node,
		  s_int index_parent,
		  s_int index_child)
{
 s_int	leveln;

 leveln = GetANodeLevel(child_node);

 return condp[index_parent][index_child][leveln];
} //end of TSBN::GetCP() 



///////////////////////////////////////////////////////////
//
//  void TSBN::PostRootPiMessage()
//  Description:
//  To Post the pi messages from root to its childrn.
//
///////////////////////////////////////////////////////////
void TSBN::PostRootPiMessage()
{
 // lambda-values = prior probilities
 GetPiValue( 0 );
 UpdatePiMessage(0);
} // end of PostRootPiMessage()  

          


///////////////////////////////////////////////////////////
//
//  void TSBN::PostRootPiMessageBest()
//  Description:
//  To Post the best pi messages from root to its childrn.
//   
///////////////////////////////////////////////////////////
void TSBN::PostRootPiMessageBest()
{
 // lambda-values = prior probilities
 GetPiValueBest( 0 );
 UpdatePiMessage(0);
} // end of PostRootPiMessageBest()  



///////////////////////////////////////////////////////////
//
// TSBN::UpdateLambdaMessage(int node)
// Description:
// To update the lambda-messge of node NODE sending to
// its parent.
// Note: Only one parent case is considered.
// The following variable is involved:
// tree[node].lambda_message,
// rlambda,
// nitemlambda.
//
///////////////////////////////////////////////////////////
void TSBN::UpdateLambdaMessage(int node)
{
 s_int	i, j;

 if ( tree[node].parent == -1 ) return; //no parent
 // Only one parent case is considered
 for (i=0; i<num_classes; ++i) {
     tree[node].lambda_message[i] 
         = condp[tree[node].condp_index][i][0] * 
           tree[node].lambda_value[0];
     for (j=1; j<num_classes; ++j) {
         tree[node].lambda_message[i] 
             += condp[tree[node].condp_index][i][j] * 
                tree[node].lambda_value[j];
     }
 }
 Push( rlambda, &nitemlambda, tree[node].parent);
} // end of update_lambda_messag()



///////////////////////////////////////////////////////////
//
// TSBN::UpdateLambdaMessageBest(int node)
// Description:
// To update the best lambda-messge of node NODE sending to
// its parent. Note: Only one parent case is considered.
//
// The following variable is involved:
// tree[node].lambda_message,
// rlambda,
// nitemlambda.
//
///////////////////////////////////////////////////////////
void TSBN::UpdateLambdaMessageBest(int node)
{
 s_int	i, j;
 float          temp;
 if ( tree[node].parent == -1 ) return; //no parent
 // Only one parent case is considered
 for (i=0; i<num_classes; ++i) {
     tree[node].lambda_message[i] 
         = condp[tree[node].condp_index][i][0] * 
           tree[node].lambda_value[0];
     for (j=1; j<num_classes; ++j) {
         temp = condp[tree[node].condp_index][i][j] * 
                tree[node].lambda_value[j];
         if (temp > tree[node].lambda_message[i])
            tree[node].lambda_message[i] = temp;
     }
 }
 Push( rlambda, &nitemlambda, tree[node].parent);

} // end of UpdateLambdaMessageBest()


///////////////////////////////////////////////////////////
//
// void TSBN::GetLambdaValue(int NODE)
// Description:
// To update the scaled lambda-value of node NODE. Missing
// data are considered by setting the scaled lambda-values 
// of the corresponding leaf nodes to 1/num_classes and the
// scaled parameter to num_classes.
//
// Variables involved:
// tree[NODE].lambda_value,
// scale_d.
//
//////////////////////////////////////////////////////////
void TSBN::GetLambdaValue(int node) 
{
 short  int	i, j, nclass1=num_classes-1;
 int		child_node = tree[node].child1;
 int  		e_node;
 float          scale = 0; // For scaling
 float          eqp = 1.0/num_classes; // equal prob.

 if ( tree[node].child1 == -1 ) {// leaf nodes
    e_node = node -  sum_nodes_levels[num_levels-1];
    switch ( evidence_type ) {
    case VIRTUAL:
     	 for (i=0; i<num_classes; ++i) {
             tree[node].lambda_value[i] = 
             virtual_evidence[e_node][i];
             scale += tree[node].lambda_value[i];
         }
         if ( fabs(scale) > ZERO ) {
            scale_d += (float)log((double)scale);
	    ScaleArray( tree[node].lambda_value, scale ); 
         }
         else {
           scale_d += (float)log((double)num_classes); 
	   for (i=0; i<num_classes; ++i) {
               tree[node].lambda_value[i] = eqp;
           }
	 }
         break;
    case HARD:
         // If a value is out of the range of num_class,
         // then the node is treated as a node with 
         // hidden value. Set all lambda_values to 1. 
         // But since scalling procdure is working, 
         // we set all lambda_values to 1/nclass and leave
         // log(scale) = log(num_classes).
         if ((hard_evidence[e_node]<0) || 
             (hard_evidence[e_node]>nclass1)) {
            scale_d += (float)log((double)num_classes);
            for (i=0; i<num_classes; ++i) 
                tree[node].lambda_value[i] = eqp;
         }
         else // observed values     
            for (i=0; i<num_classes; ++i) {
                if ( i == hard_evidence[e_node] )
                   tree[node].lambda_value[i] = 1;
                else
                   tree[node].lambda_value[i] = 0;
            }
         break;
    }
 }
 else {
    s_int le;
    le = GetANodeLevel(node); 
    for (i=0; i<num_classes; ++i) {
        tree[node].lambda_value[i] = 
        tree[child_node].lambda_message[i];
        for (j=1; j<num_children[le]; ++j) 
            tree[node].lambda_value[i] *= 
            tree[child_node+j].lambda_message[i]; 
        scale += tree[node].lambda_value[i];  
    }
    if ( fabs(scale) > ZERO ) {
       scale_d += (float)log((double)scale);
       ScaleArray( tree[node].lambda_value, scale ); 
    }
    else 
       for (i=0; i<num_classes; ++i)
           tree[node].lambda_value[i] = eqp;
 } 
} // end of  TSBN::GetLambdaValue()



///////////////////////////////////////////////////////////
//
// void TSBN::GetLambdaValueUnscaled()
//
// Description:
// To update the  lambda-value of node NODE. Missing
// data are considered by setting the  lambda-values 
// of the corresponding leaf nodes to 1  and the
//
// Variables involved:
// tree[NODE].lambda_value.
//
//////////////////////////////////////////////////////////
float* TSBN::GetLambdaValueUnscaled(int node) 
{
 short  int	i, j;
 int		child_node = tree[node].child1;
 int  		e_node;
 float* 	lambda_value;

 lambda_value = new float[num_classes];
 if ( tree[node].child1 == -1 ) {// leaf nodes
    e_node = node -  sum_nodes_levels[num_levels-1];
    switch ( evidence_type ) {
    case VIRTUAL:
     	 for (i=0; i<num_classes; ++i) {
             lambda_value[i] = virtual_evidence[e_node][i];
         }
         break;
    case HARD:
         for (i=0; i<num_classes; ++i) {
             if ( i == hard_evidence[e_node] )
                lambda_value[i] = 1;
             else
                lambda_value[i] = 0;
         }
         break;
    }
 }
 else {
    s_int le;
    le = GetANodeLevel( node ); // 06/01/98
    for (i=0; i<num_classes; ++i) {
        lambda_value[i] = tree[child_node].lambda_message[i];
        for (j=1; j<num_children[le]; ++j) 
            lambda_value[i] *= 
            tree[child_node+j].lambda_message[i]; 
    }
 } 

 return lambda_value;
} // end of  TSBN::GetLambdaValueUnscaled()



///////////////////////////////////////////////////////////
//
// float* TSBN::ScaleArray( float* array, float scale)
// Description:
// To scale an ARRAY by SCALE
//
///////////////////////////////////////////////////////////
void TSBN::ScaleArray( float* array, 
                          float scale )
{
 int 	i;
 if ( fabs(scale) >ZERO  ) {
    for (i=0; i<num_classes; ++i)
         array[i] /= scale;
 }
 
} // end of ScaleArray()


///////////////////////////////////////////////////////////
//
// s_int TSBN::GetANodeLevel(int NODE)
// Description:
// To find the level index where the node NODE is located 
// in the TSBN.
//
//////////////////////////////////////////////////////////
s_int TSBN::GetANodeLevel(int node)
{ 
 s_int	i=0;

 while ( (node >= sum_nodes_levels[i]) && 
         (i < num_levels) ) 
        ++i;
 i--;
 return i;
} 



///////////////////////////////////////////////////////////
//
// void TSBN::UpdatePiMessage( int NODE )
// Description:
// To update the pi-messages received by node NODE's.
// children from NODE.
//
// Involved variables:
// tree[child_node].pi_message,
// rpi,
// nitempi.
//
///////////////////////////////////////////////////////////
void TSBN::UpdatePiMessage( int node )
{
 s_int	        i, j, childn, le;
 int		child_node, child_node1;
 float 		pi_m;
 float          scale;

 child_node1 = tree[node].child1;
 le = GetANodeLevel( node );

 for (childn=0; childn<num_children[le]; ++childn) {
     child_node = child_node1 + childn;
     scale = 0;
     for (i=0; i<num_classes; ++i) {
         pi_m = tree[node].pi_value[i];
         for (j=0; j<num_children[le]; ++j)
             if (j != childn) 
                pi_m *=tree[child_node1+j].lambda_message[i];
         tree[child_node].pi_message[i] = pi_m;
         scale += pi_m;
     }
     if ( scale > ZERO ) {
        for (i=0; i<num_classes; ++i) 
            tree[child_node].pi_message[i] /= scale;   
     }  
     // put nodes receiving pi-message to stock
     Push( rpi, &nitempi, child_node);
 }
} // end of  TSBN::UpdatePiMessage()



////////////////////////////////////////////////////////////
//
// void TSBN::Push(int* stock, 
//                 int* nitem, 
//                 int node)
// Description:
// To put ut a node into stack if the node is a new node
// otherwise ignore it.
// 
////////////////////////////////////////////////////////////
void TSBN::Push(int*	stock, 
                int*	nitem, 
                int	node)
{
 s_int	flag=1;
 int	i;

 if (*nitem >= nitemstack) {
     cout << "The stack is too small.\n";
     exit(1);
 }

 for (i=0; i<*nitem; ++i) {
     if (stock[i] == node) {
         flag = 0;
         break;
     }
 }
 if ( flag ) {
     stock[*nitem] = node;
     *nitem = *nitem + 1;
 }
} // end of   TSBN::Push()


////////////////////////////////////////////////////////////
//
// int  TSBN::Pop(int* stock, 
//                int* nitem)
// Description:
// To Pop the first  node out of stock
//
///////////////////////////////////////////////////////////
int  TSBN::Pop(int* stock, int* nitem)
{
 int 	i, node=stock[0];

 for (i=1; i<*nitem; ++i)
 stock[i-1] = stock[i];

 *nitem = *nitem - 1;
 return node;
} // end of  TSBN::Pop


/////////////////////////////////////////////////////////////
//
//  void TSBN::PostLeafLambdaMessage()
//  Description:
//  To post the lambda messages from leaves  to its parents.
// 
//  Involved variables:
//  tree[leaf nodes].lambda_value,
//  tree[leaf nodes].lambda_message.
//
/////////////////////////////////////////////////////////////
void TSBN::PostLeafLambdaMessage()
{
 int	i, n, nn, level1, n1;

 level1 = num_levels - 1;
 n = nodes_levels[level1];
 nn = sum_nodes_levels[level1];
 for (i=0; i<n; ++i) {
     n1 = nn+i;
     GetLambdaValue( n1 );
     UpdateLambdaMessage( n1);
 }
}  // end of PostLeafLambdaMessage()  

 
/////////////////////////////////////////////////////////////
//
//  void TSBN::PostLeafLambdaMessageBest()
//  Description:
//  To post the best lambda messages from leaves  to its 
//  parents.
// 
//  Involved variables:
//  tree[leaf nodes].lambda_value,
//  tree[leaf nodes].lambda_message.
//
/////////////////////////////////////////////////////////////
void TSBN::PostLeafLambdaMessageBest()
{
 int	i, n, nn, level1, n1;

 level1 = num_levels - 1;
 n = nodes_levels[level1];
 nn = sum_nodes_levels[level1];
 for (i=0; i<n; ++i) {
     n1 = nn+i;
     GetLambdaValue( n1 );
     // UpdateLambdaMessage( (nn+i), virtual_evidence[i] );
     UpdateLambdaMessageBest( n1);
 }
}  // end of PostLeafLambdaMessageBest()  

 

///////////////////////////////////////////////////////////
// 
// void TSBN::PropagateB()
// Description:
// Lambda-message passing (bottom to top).
//
// Involved variables:
// tree[NODEs in TSBN].lambda_value,
// tree[NODEs in TSBT].lambda_message.
//
//////////////////////////////////////////////////////////
void TSBN::PropagateB()
{
 int 	        node;
 
 while  ( nitemlambda > 0) {
        node = Pop(rlambda, &nitemlambda);
        if ( tree[node].parent != -1 ) { // not root node
            GetLambdaValue( node );
            UpdateLambdaMessage( node );
        }
        else GetLambdaValue( node );// root node
 } // end of while
} // end of TSBN::PropagateB




//////////////////////////////////////////////////////////
//
// void TSBN::PropagateBBest()
// Description:
// Best lambda-message passing (bottom to top).
//
// Involved variables:
// tree[NODEs in TSBN].lambda_value,
// tree[NODEs in TSBT].lambda_message.
//
//////////////////////////////////////////////////////////
void TSBN::PropagateBBest()
{
 int 	        node;
 
 while  ( nitemlambda > 0) {
        node = Pop(rlambda, &nitemlambda);
        if ( tree[node].parent != -1 ) { // not root node
            GetLambdaValue( node );
            UpdateLambdaMessageBest( node );
        }
        else GetLambdaValue( node );// root node
 } // end of while
} // end of TSBN::PropagateBBest




///////////////////////////////////////////////////////////
//
//  void  TSBN::PropagateC()
//  Description:
//  Pi-message passing (top to  bottom).
//
//  Involved variables:
//  tree[NODEs in TSBN].pi_value,
//  tree[NODEs in TSBT].pi_message.
//           
//////////////////////////////////////////////////////////
void  TSBN::PropagateC()
{
 int		node;

 while ( nitempi > 0 ) {
       node = Pop(rpi, &nitempi);
       if ( tree[node].child1 != -1 ) { // no leaf nodes
          // post a new pi-message to its children
          GetPiValue(  node );
          UpdatePiMessage( node );
       }
       else GetPiValue(  node );// leaf nodes
 } // end of while
} // end of  TSBN::PropagateC() 



///////////////////////////////////////////////////////////
//
//  void  TSBN::PropagateCBest()
//  Description:
//  Best pi-message passing (top to  bottom).
//
//  Involved variables:
//  tree[NODEs in TSBN].pi_value,
//  tree[NODEs in TSBT].pi_message.
//
//////////////////////////////////////////////////////////
void  TSBN::PropagateCBest()
{
 int		node;

 while ( nitempi > 0 ) {
       node = Pop(rpi, &nitempi);
       if ( tree[node].child1 != -1 ) { // no leaf nodes
          // post a new pi-message to its children
          GetPiValueBest(  node );
          UpdatePiMessage( node );
       }
       else GetPiValueBest(  node );// leaf nodes
 } // end of while
} // end of  TSBN::PropagateCBest() 






//////////////////////////////////////////////////////////
//
// void  TSBN::Output(char* FNAME)
// Description:
// To ouput the segmentated label images for all levels 
// in the TSBN. The segmentated label images are stored
// in files FNAME.LEVELi, for i=0,...(num_levels-1) 
// indicating the levels.
// ( Called by ...)
//
/////////////////////////////////////////////////////////
void  TSBN::Output(char* fname)
{
 char 	*fname1=new char[strlen(fname)+3];
 s_int	leveln, row, col, hrow, hcol, inrow, incol;
 s_int	i, i1, im, j, j1, jm, table[2][2]={0,1,2,3};
 int	node, i1hcol;
 
 leveln = 0;
 sprintf(fname1, "%s.LEVEL%d", fname, leveln);
 ofstream out(fname1);
 out << GetBestClass(0)+1 << " ";
 out << "\n";
 out.close();

 // somethings worong here -- 10/01/98
 for (leveln=1; leveln<num_levels; ++leveln) {
     sprintf(fname1, "%s.LEVEL%d", fname, leveln);
     ofstream out(fname1);
     //row = sqrt((double)nodes_levels[leveln]);
     row = num_rows[leveln];
     col = num_cols[leveln];
     hrow = num_rows[leveln-1];
     hcol = num_cols[leveln-1];
     for (i=0; i<row; ++i) {
         inrow = row/hrow;
         i1 = i / inrow;
         im = i - i1*inrow;
         i1hcol = i1*hcol;
         for (j=0; j<col; ++j) {
             incol = col/hcol;
             j1 = j / incol;
             jm = j - j1*incol;
             node = (i1hcol+j1)*num_children[leveln-1] + 
                    table[im][jm] + 
                    sum_nodes_levels[leveln];
             // Node: plus 1 to consist with the labels
             out << GetBestClass(node)+1 << " ";
         }
         out << "\n";
     }
     out.close();
 }
} // end of TSBN::output()




//////////////////////////////////////////////////////////
//
// void  TSBN::Output(double* labelB)
// Description:
// To output the segmented label image at leaf level
// in TSBN.
//
/////////////////////////////////////////////////////////
void  TSBN::Output(double* labelB)
{
 s_int	leveln, row, col, hrow, hcol, inrow, incol;
 s_int	i, i1, im, j, j1, jm, table[2][2]={0,1,2,3};
 int	node, i1hcol;
 
 leveln = NUMBER_CLASSES;
 row = num_rows[leveln];
 col = num_cols[leveln];
 hrow = num_rows[leveln-1];
 hcol = num_cols[leveln-1];
 for (i=0; i<row; ++i) {
     inrow = row/hrow;
     i1 = i / inrow;
     im = i - i1*inrow;
     i1hcol = i1*hcol;
     for (j=0; j<col; ++j) {
         incol = col/hcol;
         j1 = j / incol;
         jm = j - j1*incol;
         node = (i1hcol+j1)*num_children[leveln-1] + 
                table[im][jm] + sum_nodes_levels[leveln];
         // Node: plus 1 to consist with the labels
         labelB[j*row+i] = GetBestClass(node)+1;
     }
 }
} // end of TSBN::Output()





//////////////////////////////////////////////////////////
//
// void  TSBN::Output(s_int*  LABELS1, 
//                    float** CONF )
// Description:
// To calculate the confusion matrix for a single image. 
// LABELS1: A ground true labeled image,
// CONF:    Confusion matrix.
// 
// See also
// ClassifyTSBN::ClassifyBest(char*    fimage1, 
//                            char*    fimage2, 
//                            float*** probs,   
//                            char*    fname)
// ClassifyTSBN::Classify(char*    fimage1, 
//                        char*    fimage2, 
//                        float*** probs,   
//                        char*    fname)
//
/////////////////////////////////////////////////////////
void  TSBN::Output(s_int* labels1, float** conf)
{
 int	        i,  i1, i2;

 i1 = sum_nodes_levels[num_levels-1];
 i2 = nodes_levels[num_levels-1];
 for (i=0; i<i2; ++i) 
     ++conf[labels1[i]+1][(int)GetBestClass(i1+i)+1]; 
} // end of TSBN::Output()


//////////////////////////////////////////////////////////
//
// void  TSBN::OutputEntropy(char* FNAME)
// Description:
// To calculate the entropy at each pixel and output
// a nrow*ncol matrix with each element being the 
// entropy of this pixel.  The entropy is calculated 
// based on the posterior probability.
// FNAME -- Output file for storing the entropy.
//
/////////////////////////////////////////////////////////
void  TSBN::OutputEntropy(char* fname)
{
 s_int	leveln, row, col, hrow, hcol, inrow, incol;
 s_int	i, i1, im, j, j1, jm, table[2][2]={0,1,2,3};
 int	node, i1hcol;
 
 ofstream out(fname);
 leveln = num_levels-1;
 row = num_rows[leveln];
 col = num_cols[leveln];
 hrow = num_rows[leveln-1];
 hcol = num_cols[leveln-1];
 for (i=0; i<row; ++i) {
     inrow = row/hrow;
     i1 = i / inrow;
     im = i - i1*inrow;
     i1hcol = i1*hcol;
     for (j=0; j<col; ++j) {
         incol = col/hcol;
         j1 = j / incol;
         jm = j - j1*incol;
         node = (i1hcol+j1)*num_children[leveln-1] + 
                 table[im][jm] + sum_nodes_levels[leveln];
         out << CalEntropy(node) << " ";
     }
     out << "\n";
 }
 out.close();
} // end of TSBN::outputEntropy()


//////////////////////////////////////////////////////////
//
// float  TSBN::CalEntropy(int node)
// Description:
// To calculate and return the entropy of NODE.
// The entropy is calculated based on the posterior
// probability. Entropy unit: bit
// 
/////////////////////////////////////////////////////////
float  TSBN::CalEntropy(int node) 
{
 s_int		i;
 float          bel, tol=0.0, entropy=0.0;
 float*   	bels = new float[num_classes];

 for (i=0; i<num_classes; ++i) {
        bels[i] = tree[node].lambda_value[i] * 
                  tree[node].pi_value[i];    
        tol += bels[i];
 }  
        
 for (i=0; i<num_classes; ++i) {
     bel = bels[i] / tol;
     if (bel > ZERO)
        entropy -= bel*log((double)bel)*1.4427;
 }

 return entropy;
} // end of TSBN::calEntropy()




//////////////////////////////////////////////////////////
//
//  void TSBN::GetPiValue( int NODE)
//  Description:
//  To calculate the pi-values of node NODE.
//
//  Involved variables:
//  tree[NODE].pi_value
//
//////////////////////////////////////////////////////////
void  TSBN::GetPiValue( int node)
{
 s_int	i, j;

 if ( tree[node].parent == -1 ) { // root node
    for (i=0; i<num_classes; ++i) 
         tree[node].pi_value[i] = prior_prob[i];
 }
 else {
     for (j=0; j<num_classes; ++j) { 
         tree[node].pi_value[j] = 
         tree[node].pi_message[0] * 
         condp[tree[node].condp_index][0][j];
         for (i=1; i<num_classes; ++i) {
             tree[node].pi_value[j] += 
             tree[node].pi_message[i] * 
             condp[tree[node].condp_index][i][j];
         }
     }
 }
} // end of TSBN::GetPiValue()





///////////////////////////////////////////////////////////
//
//  void TSBN::GetPiValueBest( int node)
//  Description:
//  To calculate the best pi-values of node NODE.
//
//  Involved variables:
//  tree[NODE].pi_value
//
///////////////////////////////////////////////////////////
void  TSBN::GetPiValueBest( int node)
{
 s_int		i, j;
 float          temp;

 if ( tree[node].parent == -1 ) { // root node
    for (i=0; i<num_classes; ++i) 
         tree[node].pi_value[i] = prior_prob[i];
 }
 else {
     for (j=0; j<num_classes; ++j) { 
         tree[node].pi_value[j] = 
         tree[node].pi_message[0] * 
         condp[tree[node].condp_index][0][j];
         for (i=1; i<num_classes; ++i) {
             temp = tree[node].pi_message[i] * 
                    condp[tree[node].condp_index][i][j];
             if (temp > tree[node].pi_value[j])
                tree[node].pi_value[j] = temp;
         }
     }
 }
} // end of TSBN::GetPiValueBest()






///////////////////////////////////////////////////////////
//
// s_int  TSBN::GetBestClass(int NODE)
// Description:
// To classify the pixel corresponding to node NODE by 
// using the Bayes' decision rule. 
//
///////////////////////////////////////////////////////////
s_int TSBN::GetBestClass(int node)
{
 s_int		classn, best_class;
 float*		bels;

 bels = GetBEL(node);
 best_class = 0;
 for (classn=1; classn<num_classes; ++classn) 
     if ( bels[best_class] < bels[classn] ) 
        best_class = classn;
     
 // if (bels[best_class]<0.5) best_class=-1;
 delete [] bels;

 return best_class;
} // end of TSBN::GetBestClass()


///////////////////////////////////////////////////////////
//
// float*  TSBN::GetBEL(int NODE)
// Description:
// To calculate the belief(BEL) at node NODE.
//
///////////////////////////////////////////////////////////
float*  TSBN::GetBEL(int node)
{
 s_int		classn;
 float*         bels = new float[num_classes];

 for (classn=0; classn<num_classes; ++classn)
     bels[classn] = tree[node].lambda_value[classn] * 
                    tree[node].pi_value[classn];
 return bels;
} // end of TSBN::GetBEL()


///////////////////////////////////////////////////////////
//
// void TSBN::InitialLambdaPi()
// Description:
// To initialise the lambda-message and pi-message for
// each node in the TSBN.
// 
// Involved variables:
// tree[NODES in TSBN].lambda_message,
// tree[NODES in TSBN].pi_message.
//
////////////////////////////////////////////////////////////
void TSBN::InitialLambdaPi()
{
 int 	i, n;

 for (n=0; n<num_nodes; ++n)
 for (i=0; i<num_classes; ++i) {
     tree[n].lambda_message[i] = 0;
     //tree[n].lambda_value[i] = 0;
     tree[n].pi_message[i] = 0;
 }
} // end TSBN::InitialLambdaPi()



///////////////////////////////////////////////////////////
//
// float TSBN::Loglikelihood( )
// Description:
// To calculate the log(P(Evidences)).
//
//
///////////////////////////////////////////////////////////
float  TSBN::Loglikelihood(  )
{
 int     i;
 float   p=0;

 for (i=0; i<num_classes; ++i)
     p += tree[0].lambda_value[i]*tree[0].pi_value[i];

 scale_d = log(p) + scale_d;
 return scale_d;
} // end of Loglikelihood


//////////////////////////////////////////////////////////
//
// void TSBN::EstimateCPT(float*** PROBS_OUT)
// Description:
// To calculate the estimated CPTs. There is no constraint 
// within CPT.
//
//
//////////////////////////////////////////////////////////
void TSBN::EstimateCPT( float*** probs_out )
{
 s_int		i, j, k;
 double		den = 0;
 double**       alink;
 double**       dmatrix(int, int);

 // initialise 
 for (k=0; k<num_cpts; ++k)
 for (i=0; i<num_classes; ++i)
 for (j=0; j<num_classes; ++j)
     probs_out[k][i][j] = 0;

 // allocation
 alink = dmatrix(num_classes, num_classes);

 // estimate the priors
 for (k=0; k<num_classes; ++k) {
     probs_out[0][0][k] = 
     tree[0].lambda_value[k] * tree[0].pi_value[k];
     den += probs_out[0][0][k];
 }
 for (k=0; k<num_classes; ++k) 
     probs_out[0][0][k] /= den;

 // estimate the conditional probabilities
 int       nodec;
 s_int     i1, i2, ki, kj, le, nchild;
 double    temp, temp0;

 for (le=1; le<num_levels; ++le) {
     // go through links one by one
     i1 = sum_nodes_levels[le-1];
     i2 = sum_nodes_levels[le];
     //
     for (i=i1; i<i2; ++i) {
         for (j=0; j<num_children[le-1]; ++j) {
             den = 0;
             nodec = tree[i].child1 + j;
             // update the probs_out using this link
             for (ki=0; ki<num_classes; ++ki) {
	         // pi-value and lambda-message from 
	         // other children
                 temp0 = tree[i].pi_value[ki];
	         for (nchild=0; nchild<num_children[le-1]; 
                      ++nchild) {
                     if ( nchild != j ) 
                        temp0 *= 
                        tree[tree[i].child1+nchild].
                        lambda_message[ki];
                 }
                 // lambda-value and conditional probs
  	         for (kj=0; kj<num_classes; ++kj) {
		     temp = 
                     tree[nodec].lambda_value[kj]*
                     condp[tree[nodec].condp_index][ki][kj]*
                     temp0;
                     alink[ki][kj] = temp;
                     den += temp;
	         }
             }
             // when i=2049 j=2 , den =0;
             // how to explain this case: the joint prob. 
	     // table are all zero??
             // 26/01/96 ??
             //  THINK...........


	     // normalise probs on a link
             // joint probability table
             if (den > 0.0)
                for (ki=0; ki<num_classes; ++ki)
	        for (kj=0; kj<num_classes; ++kj) {
                    alink[ki][kj] /= den;
                    probs_out[tree[nodec].
                    condp_index][ki][kj] += 
                   (float)alink[ki][kj];
	        }
         }
     }
 } // end of for le
 delete [] alink;
} // end of TSBN::EstimateCPT()

 	




///////////////////////////////////////////////////////////
//
// void TSBN::EstimateCPTConstrained(float***)
// Description:
// To calculate the estimated CTPS.  The CPT is shared in 
// every level. There is constraint within CPT: off-diagonal 
// terms are equal
//
//
///////////////////////////////////////////////////////////
void TSBN::EstimateCPTConstrained( float*** probs_out )
{
 s_int		i, j, k;
 float		den = 0;
 float**        alink;

 // allocation
 alink = fmatrix(num_classes, num_classes);

 // estimate the priors
 for (k=0; k<num_classes; ++k) {
     probs_out[0][k][0] = tree[0].lambda_value[k]* 
                          tree[0].pi_value[k];
     den += probs_out[0][k][0];
 }
 for (k=0; k<num_classes; ++k) 
     probs_out[0][k][0] /= den;

 // estimate the conditional probabilities
 s_int	   i1, i2, ki, kj, le, nchild;
 int       nodec;
 int       nlinks;
 float     temp, temp0;
 for (le=1; le<num_levels; ++le) {
     // initialise
     for (ki=0; ki<num_classes; ++ki)
     for (kj=0; kj<num_classes; ++kj)
         probs_out[ki][kj][le] = 0;
     // go through links one by one
     i1 = sum_nodes_levels[le-1];
     i2 = sum_nodes_levels[le];
     //
     nlinks = 0;
     for (i=i1; i<i2; ++i) {
         for (j=0; j<num_children[le-1]; ++j) {
             den = 0;
             nodec = tree[i].child1 + j;
             // update the probs_out using this link
             for (ki=0; ki<num_classes; ++ki) {
	         // pi-value and lambda-message from
	         // other children
                 temp0 = tree[i].pi_value[ki];
	         for (nchild=0; nchild<num_children[le-1]; 
                      ++nchild) {
                     if ( nchild != j ) 
                        temp0 *= 
                        tree[tree[i].child1+nchild].
                        lambda_message[ki];
                 }
                 // lambda-value and conditional probs
  	         for (kj=0; kj<num_classes; ++kj) {
		     temp = 
                     tree[nodec].lambda_value[kj]*
                     condp[tree[nodec].condp_index][ki][kj]*
                     temp0;
                     alink[ki][kj] = temp;
                     den += temp;
	         }
             }
	     // normalise probs on a link
             for (ki=0; ki<num_classes; ++ki)
	     for (kj=0; kj<num_classes; ++kj) 
                 alink[ki][kj] /= den;
	     // shared elements in CPT
             temp0 = 0;
             temp = 0;
             for (ki=0; ki<num_classes; ++ki)
	     for (kj=0; kj<num_classes; ++kj) 
                 if (ki==kj)
                    temp0 += alink[ki][kj];
                 else
                    temp += alink[ki][kj];
             temp /= num_classes - 1;
             // asign the estimated shared elements
             for (ki=0; ki<num_classes; ++ki)
	     for (kj=0; kj<num_classes; ++kj) 
                 if (ki==kj)
                    probs_out[ki][kj][le] += temp0;
                 else
                    probs_out[ki][kj][le] += temp;   
         }
     }
 } // end of forr le
} // end of TSBN::EstimateCPTConstrained()

 	



///////////////////////////////////////////////////////////
//
// void TSBN::MatchLabel(float*** probs_out)
// Description:
// To match the labels of hidden nodes to the oserved nodes.
//  
//////////////////////////////////////////////////////////
void TSBN::MatchLabel( float*** probs_out )
{
 s_int	        i, j, le;
 s_int*         index = new s_int [num_classes];
 float          ruler, temp;
 float*         temp_row = new float [num_classes];

 for (le=(num_levels-1); le>0; --le) {
     // find the index

     // 0 1 2 
     ruler = probs_out[0][0][le]+ 
             probs_out[1][1][le]+ 
             probs_out[2][2][le];
     index[0] = 0; index[1] = 1; index[2] = 2;
     // 0 2 1
     temp = probs_out[0][0][le]+ 
            probs_out[1][2][le]+ 
            probs_out[2][1][le];
     if (temp > ruler) {
         index[0] = 0; index[1] = 2; index[2] = 1;
         ruler = temp;
     }
     // 1 0 2
     temp = probs_out[0][1][le]+ 
            probs_out[1][0][le]+ 
            probs_out[2][2][le];
     if (temp > ruler) {
         index[0] = 1; index[1] = 0; index[2] = 2;
         ruler = temp;
     }
     // 1 2 0
     temp  = probs_out[0][1][le]+ 
             probs_out[1][2][le]+ 
             probs_out[2][0][le];
     if (temp > ruler) {
        index[0] = 1; index[1] = 2; index[2] = 0;
        ruler = temp;
     }
     // 2 0 1
     temp = probs_out[0][2][le]+ 
            probs_out[1][0][le]+ 
            probs_out[2][1][le];
     if (temp > ruler) {
        index[0] = 2; index[1] = 0; index[2] = 1;
        ruler = temp;
     }
     // 2 1 0
     temp = probs_out[0][2][le]+ 
            probs_out[1][1][le]+ 
            probs_out[2][0][le];
     if (temp > ruler) {
        index[0] = 2; index[1] = 1; index[2] = 0;
        ruler = temp;
     }
     // switch the rows 
     for (j=0; j<num_classes; ++j )
         temp_row[j] = probs_out[index[0]][j][le];
     for (i=1; i<num_classes; ++i) 
     for (j=0; j<num_classes; ++j)
         probs_out[index[i-1]][j][le] =  
         probs_out[index[i]][j][le];
     for (j=0; j<num_classes; ++j )
         probs_out[index[num_classes-1]][j][le] = 
         temp_row[j];  
 }  

} // end of TSBN::MatchLabel()


///////////////////////////////////////////////////////////
//
// void TSBN::SetEvidence( s_int* LABELS)
// Description:
// In the case that hard-evidence are used in trining,
// set the evidence type to HARD and the pointer of 
// hard_evidence to the labels LABLES.
// 
// Involved variables:
// hard_evidence,
// evidence_type.
// 
///////////////////////////////////////////////////////////
void TSBN::SetEvidence( s_int* labels )
{
  hard_evidence = labels;
  evidence_type = HARD;
}


////////////////////////////////////////////////////////////
//
// void TSBN::SetVirtualEvidence( float** EV )
// Description:
// In the case that virtual-evidence are used in trining,
// set the evidence type to VIRTUAL and the pointer of 
// virtual_evidence to the evidence EV.
// 
// Involved variables:
// virtual_evidence,
// evidence_type.
// 
////////////////////////////////////////////////////////////
void TSBN::SetVirtualEvidence( float** ev )
{ 
  virtual_evidence = ev;
  evidence_type = VIRTUAL;
}



///////////////////////////////////////////////////////////
//
// void TSBN::SetInitialValue(  )
// Description:
// To set the scaling factor to zero.
//  
///////////////////////////////////////////////////////////
void TSBN::SetInitialValue( )
{
  scale_d = 0;
}




//////////////////////////////////////////////////////////
//
// void  output_prior(char* FNAME)
// Modifications:
//	19/06/97
//////////////////////////////////////////////////////////
/*
void  TSBN::output_prior(char* fname)
{
 char*	fname1=new char[strlen(fname)+3];
 s_int	leveln, row, col, hrow, hcol, inrow, incol;
 s_int	i, i1, im, j, j1, jm, k, table[2][2]={0,1,2,3};
 int	node, i1hcol;
 
 printf("output\n");
 leveln = 0;
 sprintf(fname1, "%s.LEVEL%d", fname, leveln);
 ofstream out(fname1);
 out << GetBestClass(0)+1 << " ";
 out << "\n";
 out.close();

 // somethings worong here -- 10/01/98
 for (leveln=1; leveln<num_levels; ++leveln) {
     sprintf(fname1, "%s.LEVEL%d", fname, leveln);
     ofstream out(fname1);
     //row = sqrt((double)nodes_levels[leveln]);
     row = num_rows[leveln];
     col = num_cols[leveln];
     hrow = num_rows[leveln-1];
     hcol = num_cols[leveln-1];
     for (i=0; i<row; ++i) {
         inrow = row/hrow;
         i1 = i / inrow;
         im = i - i1*inrow;
         i1hcol = i1*hcol;
         for (j=0; j<col; ++j) {
             incol = col/hcol;
             j1 = j / incol;
             jm = j - j1*incol;
             node = (i1hcol+j1)*num_children[leveln-1]+ 
                    table[im][jm] + 
                    sum_nodes_levels[leveln];
             // Node: plus 1 to consist with the labels
             for (k=0; k<num_classes; ++k)
                 out << tree[node].pi_value[k] << " ";
              out << "\n";
         }
     }
     out.close();
 }

} // end of TSBN::output()

*/

///////////////////////////////////////////////////////////
//
// float*  TSBN::Belief2VE( float** labels)
// Description:
// To obtain the best probabilities at node       
//   
//
//  ???
///////////////////////////////////////////////////////////
void  TSBN::Belief2VE(  )
{
 int		i1=sum_nodes_levels[num_levels-1];
 int            i;
 int            i2=nodes_levels[num_levels-1];
 s_int          classn;
 
 for (i=0; i<i2; ++i) {
     for (classn=0; classn<num_classes; ++classn)
         virtual_evidence[i][classn] = 
                 tree[i+i1].lambda_value[classn] * 
                 tree[i+i1].pi_value[classn];
 }
}



///////////////////////////////////////////////////////////
//
//  void  TSBN::PassMessage( )
//  Description:
//  Passing lambda-messages and pi-messages through all 
//  nodes in the TSBN.
//  
///////////////////////////////////////////////////////////
void TSBN::PassMessage( )
{
 PostLeafLambdaMessage(); 
 PropagateB();
 PostRootPiMessage(); 
 PropagateC(); 
} // end of TSBN::PassMessage() 

		    

///////////////////////////////////////////////////////////
//
//  void  TSBN::PassMessageBest( )
//  Description:
//  Passing best lambda-messages and best pi-messages 
//  through all nodes in the TSBN.
//  
///////////////////////////////////////////////////////////
void TSBN::PassMessageBest( )
{
 PostLeafLambdaMessageBest(); 
 PropagateBBest();
 PostRootPiMessageBest(); 
 PropagateCBest(); 
} // end of TSBN::PassMessageBest() 

		    


///////////////////////////////////////////////////////////
//
// void TSBN::CalculatePrior( s_int type )
// Description:
// To calculate the priors for all nodes on the trees
// TYPE: The tyep of CPTs; TYPE=6-> six different CPTS 
//       on the links between level0 and level 1;
//
// Xiaojuan Fegn 12/01/99
//
///////////////////////////////////////////////////////////
void TSBN::CalculatePrior( s_int type )
{
 s_int	        i, j;
 int		node;

 if (type !=6 ) 
    errors(" Only cosider 6 CPTs on the links between level0 \
             and level1 at this moment !\n");

 // The root node
 for (i=0; i<num_classes; ++i) {
     tree[0].prior[i] = prior_prob[i];
 }
 // Calculate the priors for nodes
 for (node=1; node<num_nodes; ++node) {
     for (i=0; i<num_classes; ++i) {
         tree[node].prior[i] = 0.0;
         for (j=0; j<num_classes; ++j) 
             tree[node].prior[i] +=
             condp[tree[node].condp_index][j][i] *
             tree[tree[node].parent].prior[j];
     }
 }
} // end of  TSBN::CalculatePrior()




///////////////////////////////////////////////////////////
// 
// void TSBN::CalculateLambdaMargProb()
// Description:
// Calculate the log Marginal probablity as Lambda-message 
// passing (bottom to top).
//
// Xiaojuan Feng  12/01/99
//
//////////////////////////////////////////////////////////
void TSBN::CalculateLambdaMargProb()
{
 int 	        node;
 
 PostLeafLambdaMessage2();

 while  ( nitemlambda > 0) {
        node = Pop(rlambda, &nitemlambda);
        if ( tree[node].parent != -1 ) { // not root node
            CalculateLambdaMargProb( node );
            UpdateLambdaMessage( node );
        }
        else CalculateLambdaMargProb( node );// root node
 } // end of while
} // end of TSBN::CalculateLambdaMargProb



/////////////////////////////////////////////////////////////
//
//  void TSBN::PostLeafLambdaMessage2()
//  Description:
//  To post the lambda messages from leaves  to its parents.
// 
//  Involved variables:
//  tree[leaf nodes].lambda_value,
//  tree[leaf nodes].lambda_message.
//
/////////////////////////////////////////////////////////////
void TSBN::PostLeafLambdaMessage2()
{
 int	i, n, nn, level1, n1;

 level1 = num_levels - 1;
 n = nodes_levels[level1];
 nn = sum_nodes_levels[level1];
 for (i=0; i<n; ++i) {
     n1 = nn+i;
     CalculateLambdaMargProb( n1 );
     UpdateLambdaMessage( n1);
 }
}  // end of PostLeafLambdaMessage2()  

 


///////////////////////////////////////////////////////////
//
// void TSBN::CalculateLambdaMargProb(int NODE)
// Description:
// To update the scaled lambda-value of node NODE;
// Obtain the log scaling factor;
// Calculate the log marginal likelihood;  Missing
// data are considered by setting the scaled lambda-values 
// of the corresponding leaf nodes to 1/num_classes and the
// scaled parameter to num_classes.
//
// Xiaojuan Feng  12/01/99
//
//////////////////////////////////////////////////////////
void TSBN::CalculateLambdaMargProb(int node) 
{
 short  int	i, j, nclass1=num_classes-1;
 int		child_node = tree[node].child1;
 int  		e_node;
 float          scale = 0; // For scaling
 float          eqp = 1.0/(float)num_classes; // equal prob.

 if ( tree[node].child1 == -1 ) {// leaf nodes
    e_node = node -  sum_nodes_levels[num_levels-1];
    switch ( evidence_type ) {
    case VIRTUAL:
     	 for (i=0; i<num_classes; ++i) {
             tree[node].lambda_value[i] = 
             virtual_evidence[e_node][i];
             scale += tree[node].lambda_value[i];
         }
         if ( fabs(scale) > ZERO ) {
            tree[node].log_scale = -log((double)scale);
	    ScaleArray( tree[node].lambda_value, scale ); 
         }
         else {
           tree[node].log_scale = -log((double)num_classes);
	   for (i=0; i<num_classes; ++i) {
               tree[node].lambda_value[i] = eqp;
           }
	 }
         break;
    case HARD:
         // If a value is out of the range of num_class,
         // then the node is treated as a node with 
         // hidden value. Set all lambda_values to 1. 
         // But since scalling procdure is working, 
         // we set all lambda_values to 1/nclass and leave
         // log(scale) = log(num_classes).
         if ((hard_evidence[e_node]<0) || 
             (hard_evidence[e_node]>nclass1)) {
            tree[node].log_scale = -log((double)num_classes);
            for (i=0; i<num_classes; ++i) 
                tree[node].lambda_value[i] = eqp;
         }
         else {// observed values    
            tree[node].log_scale = 0.0; 
            for (i=0; i<num_classes; ++i) {
                if ( i == hard_evidence[e_node] )
                   tree[node].lambda_value[i] = 1;
                else
                   tree[node].lambda_value[i] = 0;
            }
         }
         break;
    }
 }
 else {
    s_int le;
    le = GetANodeLevel(node); 
    for (i=0; i<num_classes; ++i) {
        tree[node].lambda_value[i] = 
        tree[child_node].lambda_message[i];
        for (j=1; j<num_children[le]; ++j) 
            tree[node].lambda_value[i] *= 
            tree[child_node+j].lambda_message[i]; 
        scale += tree[node].lambda_value[i];  
    }
    if ( fabs(scale) > ZERO ) {
       tree[node].log_scale = -log((double)scale);
       ScaleArray( tree[node].lambda_value, scale ); 
    }
    else {
       tree[node].log_scale = -log((double)num_classes);
       for (i=0; i<num_classes; ++i)
           tree[node].lambda_value[i] = eqp;
    }
    // obtain the log_scale
    e_node = tree[node].child1;
    for (i=0; i<num_children[le]; ++i)
        tree[node].log_scale += tree[e_node+i].log_scale;
 } 

 // Calculate log Marginal probs
 double temp=0.0;
 for (i=0; i<num_classes; ++i)
     temp += tree[node].prior[i]*
             tree[node].lambda_value[i];
 tree[node].log_marg_prob = log(temp)-tree[node].log_scale;
} // end of  TSBN::CalculateLambdaMargProb()





//////////////////////////////////////////////////////////
//
// float*  TSBN::OutputCodingCost()
// Description:
// To ouput log marginal probability  along 
// the truncated tree.
//
// Xiaojuan  Feng  13/01/99
//
/////////////////////////////////////////////////////////
float*  TSBN::OutputLogMargProb()
{
 s_int	leveln;
 int	i, node;
 
 float* log_marg_prob = new float [num_levels];

 // Level 0
 log_marg_prob[0] = tree[0].log_marg_prob;

 //
 for (leveln=1; leveln<num_levels; ++leveln) {
     log_marg_prob[leveln]=0.0;
     for (i=0; i<nodes_levels[leveln]; ++i) {
         node = sum_nodes_levels[leveln]+i;
         log_marg_prob[leveln] += tree[node].log_marg_prob;
     }
 }
 return log_marg_prob;
} // end of TSBN::OutputLogMargProb()




